

// -*- mode:c++ -*-
//
// Module CodeGen.cc
//
// Copyright (c) 2002, 2003 Raoul M. Gough
// Copyright (C) 2006       Joel Yliluoma (http://iki.fi/bisqwit/)
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// History
// =======
// 2002/ 2/11	rmg	    File creation
// 2006-10-23 bisqwit   Rewrote to avoid #include files and scope problems
//

#include "CodeGen.hh"

#include <sstream>
#include <cassert>
#include <iostream>
#include <map>

extern const char *clamp_parser_rcsid;
const char *clamp_codegen_rcsid = "$Id: $";

using namespace std;

extern std::string infn, outfn;

namespace
{
  using namespace clamp;
  
  std::map<std::string, std::string> LambdaFileContents;
  
  static unsigned CountLines(const std::string& str)
  {
    unsigned n=1;
    std::string::size_type b=0, p;
    while( (p = str.find('\n', b)) != str.npos )
    {
      ++n;
      b = p+1;
    }
    return n;
  }

  unsigned count (paramListT *pListp)
  {
    if (pListp)
      {
	return pListp->size();
      }

    else
      {
	return 0;
      }
  }
  
  void initList (std::stringstream &strm, paramListT *pListp, bool first = true)
  {
    if (pListp)
      {
	for (paramListT::const_iterator iter = pListp->begin()
	       ; iter != pListp->end()
	       ; ++iter)
	  {
	    if (iter->get()->initialiser.get())
	      {
		if (!first)
		  {
		    strm << ", ";
		  }

		strm << *iter->get()->initialiser;

		first = false;
	      }

	    else if (iter->get()->type.get())
	      {
		if (!first)
		  {
		    strm << ", ";
		  }

                // Default initialize
		strm << "(" << *iter->get()->type << ")0";

		first = false;
	      }
	  }
      }
  }

  void typeList (std::stringstream &strm, paramListT *pListp, bool first = true)
  {
    if (pListp)
      {
	for (paramListT::const_iterator iter = pListp->begin()
	       ; iter != pListp->end()
	       ; ++iter)
	  {
	    if (iter->get()->type.get())
	      {
		if (!first)
		  {
		    strm << ", ";
		  }

		strm << *iter->get()->type;

		first = false;
	      }
	  }
      }
  }
}

namespace clamp
{
  ////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////

  CodeGen::CodeGen ()
    : mStack ()
    , mLambdaCount (0)
    , mImplStream ()
    , mLambdaMap ()
    , mGeneratorStream ()
  {
  }

  CodeGen::LambdaContext::LambdaContext (std::stringstream* f
					 , int b
					 , std::auto_ptr<lambdaT> lamp)
    : mPrevFile (f)
    , mOpenBracePos (b)
    , mLamp (lamp.release())
    , mLambdaNumber (++sContextCount)
  {
  }

  CodeGen::LambdaGroup::LambdaGroup ()
    : mLambdaList ()
    , mImplStream (new std::stringstream)
  {
  }

  void implTemplateName (std::stringstream &strm, const LambdaCategory& typeId)
  {
    strm
      << "lambda_template_"
      << typeId.ccount
      << "_"
      << typeId.pcount
      << "_"
      << typeId.quals;
  }

  void tList (std::stringstream &strm, int start, int end, bool first = true)
  {
    for (int count = start; count <= end; ++count)
      {
	if (!first)
	  {
	    strm << ", ";
	  }

	strm << "T" << count;

	first = false;
      }
  }

  void implFunctionVar (std::stringstream &strm, const LambdaCategory& typeId, const char *n)
  {
    const unsigned ccount = typeId.ccount;
    const unsigned pcount = typeId.pcount;
    
    strm << "T1 (";

    if (ccount > 0)
      {
	// Implementation via pointer to member function
	implTemplateName (strm, typeId);
	strm << "::";
      }
    // else implementation via pointer to ordinary function

    strm << "*" << n << ") (";
    tList (strm, 2 + ccount, 1 + ccount + pcount);
    strm << ")";
    if(ccount > 0) strm << typeId.quals;
  }

  void implInstance (std::stringstream &strm, lambdaT const &lam)
  {
    LambdaCategory typeId;
     typeId.ccount = count (lam.context.get());
     typeId.pcount = count (lam.params.get());
     typeId.quals  = lam.qual_str.get() ? *lam.qual_str.get() : "";
    implTemplateName (strm, typeId);

    strm
      << "<"
      << *lam.ret_type;

    typeList (strm, lam.context.get(), false);
    typeList (strm, lam.params.get(), false);

    strm << "> ";
  }

  std::auto_ptr<stringT> CodeGen::lambdaTypeName (std::auto_ptr<lambdaT> lamp)
  {
    LambdaCategory typeId;
     typeId.ccount = count (lamp->context.get());
     typeId.pcount = count (lamp->params.get());
     typeId.quals  = lamp->qual_str.get() ? *lamp->qual_str.get() : "";
    lambdaGroup (typeId);
    
    // Make sure we've recorded the existance of this lambda group

    std::stringstream temp;

    implInstance (temp, *lamp);

    return std::auto_ptr<stringT> (new stringT (temp.str()));
  }

  void CodeGen::lambdaType (std::stringstream& outfile, std::auto_ptr<lambdaT> lamp)
  {
    outfile << *lambdaTypeName(lamp);
  }

  bool CodeGen::braceMatch (int pos)
  {
    return (!mStack.empty()) && (mStack.top().mOpenBracePos == pos);
  }

  std::string contextVarExpression (paramT const &ctxParam)
  {
    std::stringstream result;

    result << "mContext" << ctxParam.ctx_param_num;

    return result.str();
  }

  int CodeGen::LambdaContext::sContextCount = 0;

  stringT bodyFileName (int count)
  {
    char name[PATH_MAX];
    sprintf (name, "lambda_body_%d.clamp_h", count);
    return stringT (name);
  }

  void explicitConstructor (std::stringstream &strm, int id, const lambdaT &lam)
  {
    if (lam.onHeap)
      {
	strm << "new ";
      }
    
    bool uses_pointer = true;

    implInstance (strm, lam);
    strm << "(";

    if(uses_pointer)
    {
      strm << "&";
      implInstance (strm, lam);
      strm << "::fn" << id;
    }
    
    initList (strm, lam.context.get(), !uses_pointer);
    strm << ")";
  }

  void typenameList (std::stringstream &strm
		     , unsigned start
		     , unsigned end
		     , bool first = true)
  {
    for (unsigned count = start; count <= end; ++count)
      {
	if (!first)
	  {
	    strm << ", ";
	  }

	strm << "typename T" << count;

	first = false;
      }
  }

  void implTemplateDecl (std::stringstream &strm, unsigned templateParams)
  {
    strm
      << "template <";

    typenameList (strm, 1, templateParams);

    strm << ">";
  }

  void CodeGen::startExplicitImpl (std::stringstream &strm, const LambdaCategory& typeId)
  {
    const unsigned ccount = typeId.ccount;
    const unsigned pcount = typeId.pcount;
    
    bool uses_pointer = true;//categorysize(typeId) != 1;
    
    implTemplateDecl (strm, ccount + pcount + 1);

    strm << "\nstruct ";

    implTemplateName (strm, typeId);

    strm << "\n{\n  ";

    if(uses_pointer)
    {
        implFunctionVar (strm, typeId, "mFn"); 
        strm << ";";
    }

    for (unsigned count = 1; count <= ccount; ++count)
      {
	strm << "\n  T" << count + 1 << " mContext" << count << ";";
      }

    strm << "\n\n  typedef T1 result_type;";

    for (unsigned count = 1; count <= ccount; ++count)
      {
	strm
	  << "\n  typedef T" << count + 1
	  << " context_type_" << count
	  << ";";
      }

    for (unsigned count = 1; count <= pcount; ++count)
      {
	strm
	  << "\n  typedef T" << count + ccount + 1
	  << " argument_type_" << count
	  << ";";
      }

    // Additional typedefs in special cases for standard binder compatability
    if (pcount == 1)
      {
	strm
	  << "\n  typedef T" << ccount + 2
	  << " argument_type;";
      }

    else if (pcount == 2)
      {
	strm
	  << "\n  typedef T" << ccount + 2
	  << " first_argument_type;\n  typedef T" << ccount + 3
	  << " second_argument_type;";
      }

    strm << "\n\n  // Constructor\n  ";

    implTemplateName (strm, typeId);

    strm << " (";
    
    bool comma = false;
    if(uses_pointer)
    {
        implFunctionVar (strm, typeId, "fn");
        comma = true;
    }

    for (unsigned count = 1; count <= ccount; ++count)
      {
        if(comma) strm << ", ";
	strm << "T" << count + 1 << " c" << count;
	comma = true;
      }

    strm << ")";
    
    std::string prefix = "\n   : ";
    
    if(uses_pointer) { strm << prefix << "mFn(fn)"; prefix = ", "; }

    for (unsigned count = 1; count <= ccount; ++count)
      {
	strm << prefix << "mContext" << count << " (c" << count << ")";
	prefix = ", ";
      }

    strm
      << "\n  { }\n"
      << "\n  // Forwarding function"
      << "\n  T1 operator() (";

    for (unsigned count = 1; count <= pcount; ++count)
      {
	if (count != 1)
	  {
	    strm << ", ";
	  }

	strm << "T" << count + ccount + 1 << " p" << count;
      }

    strm << ")" << typeId.quals
         << "\n  {\n    return (";

    if (ccount > 0)
      {
	// Have closure parameters - use pointer to member syntax
	strm << "this->";
      }
    // else static function

    if(uses_pointer)
    {
        strm << "*mFn";
    }
    else
    {
        strm << findFirstFunc(typeId);
    }
    
    strm << ")(";

    for (unsigned count = 1; count <= pcount; ++count)
      {
	if (count != 1)
	  {
	    strm << ", ";
	  }

	strm << "p" << count;
      }

    strm << ");\n  }\n";
  }

  void declareMemberFn (std::stringstream &strm
			, int id
			, const LambdaCategory& typeId)
  {
    const unsigned ccount = typeId.ccount;
    const unsigned pcount = typeId.pcount;
    
    strm << "\n  ";

    if (ccount == 0)
      {
	strm << "static ";
      }

    strm << "T1 fn" << id << " (";

    tList (strm, 2 + ccount, 1 + ccount + pcount);

    strm << ")";
    
    if(ccount > 0) strm << typeId.quals;
    
    strm << ";\n";
  }

  void defineMemberFn (std::stringstream &strm
		       , int id
		       , const LambdaCategory& typeId
		       , const lambdaT &lam)
  {
    const unsigned ccount = typeId.ccount;
    const unsigned pcount = typeId.pcount;
    
    implTemplateDecl (strm, ccount + pcount + 1);

    strm << "\nT1 ";

    implTemplateName (strm, typeId);

    strm << "<";

    tList (strm, 1, 1 + ccount + pcount);

    strm << ">::fn" << id << " (";

    if (lam.params.get())
      {
	unsigned count = 1;

	for (paramListT::const_iterator iter = lam.params->begin()
	       ; iter != lam.params->end()
	       ; ++iter)
	  {
	    if (count != 1)
	      {
		strm << ", ";
	      }

	    strm << "T" << count + ccount + 1;

	    const stringT *name = iter->get()->name.get();

	    if (name)
	      {
		strm
		  << " " << *name;
	      }

	    ++count;
	  }
      }

    strm << ")";

    if(ccount > 0) strm << typeId.quals;

    strm << "\n{";

    if (lam.explicitContext && lam.context.get())
      {
	// Context parameters require renaming
	unsigned count = 1;

	for (paramListT::const_iterator iter = lam.context->begin()
	       ; iter != lam.context->end()
	       ; ++iter)
	  {
	    const stringT *name = iter->get()->name.get();

	    if (name)
	      {
		strm
		  << "\n#define " << *name
		  << " mContext"
                  << count;
	      }

	    ++count;
	  }
      }

    strm << LambdaFileContents[bodyFileName(id)];
    strm << "\n";
  #if 0
    strm << "#line " << CountLines(strm.str()) << " \"" << outfn << "\"\n";
  #endif

    if (lam.explicitContext && lam.context.get())
      {
	unsigned count = 1;

	for (paramListT::const_iterator iter = lam.context->begin()
	       ; iter != lam.context->end()
	       ; ++iter)
	  {
	    const stringT *name = iter->get()->name.get();

	    if (name)
	      {
		strm
		  << "\n#undef " << *name;
	      }

	    ++count;
	  }
      }

    strm << "\n}\n";
  }

  void implicitConstructor (std::stringstream &strm, int id, const lambdaT &lam)
  {
    strm
      << "lambda_generator_"
      << id
      << "<"
      << *lam.ret_type;

    typeList (strm, lam.params.get(), false);

    strm << ">::generate(";

    initList (strm, lam.context.get());

    strm <<")";
  }

  void implicitTemplateInstance (std::stringstream &strm
				 , const lambdaT &lam
				 , const LambdaCategory& typeId)
  {
    const unsigned ccount = typeId.ccount;
    const unsigned pcount = typeId.pcount;
    
    implTemplateName (strm, typeId);

    strm << " <T1";

    if (ccount)
      {
	paramListT::const_iterator iter = lam.context->begin();

	for (unsigned count = 1; count <= ccount; ++count)
	  {
	    strm << ", T" << count + 1;

	    if ((*iter)->byReference)
	      {
		strm << " &";
	      }

	    ++iter;
	  }
      }

    tList (strm, 2 + ccount, 1 + ccount + pcount, false);

    strm << ">";
  }

  void CodeGen::writeGenerator (std::stringstream &strm, int id, const lambdaT &lam)
  {
    LambdaCategory typeId;
     typeId.ccount = count (lam.context.get());
     typeId.pcount = count (lam.params.get());
     typeId.quals  = lam.qual_str.get() ? *lam.qual_str.get() : "";
    const unsigned ccount = typeId.ccount;
    const unsigned pcount = typeId.pcount;

    bool uses_pointer = true;//categorysize(typeId) != 1;
    
    strm << "\ntemplate <";

    typenameList (strm, 1, 1);
    typenameList (strm, 2 + ccount, 1 + ccount + pcount, false);

    strm
      << ">\nstruct lambda_generator_" << id
      << "\n{";

    if (ccount >= 1)
      {
	strm << "\n  template <";

	typenameList (strm, 2, 1 + ccount);

	strm << ">";
      }

    strm << "\n  static ";

    implicitTemplateInstance (strm, lam, typeId);

    if (lam.onHeap)
      {
	strm << " *";
      }

    strm << "\n  generate(";

    if (ccount)
      {
	paramListT::const_iterator iter = lam.context->begin();

	for (unsigned count = 1; count <= ccount; ++count)
	  {
	    if (count != 1)
	      {
		strm << ", ";
	      }

	    strm << "T" << count + 1 << " ";

            if ((*iter)->byReference)
              {
                strm << "&";   // Accept parameter by reference
              }

            strm << "p" << count;

	    ++iter;
	  }
      }

    strm << ")\n  {\n    return ";

    if (lam.onHeap)
      {
	strm << "new ";
      }

    implicitTemplateInstance (strm, lam, typeId);

    strm << " (";
    
    bool comma = false;
    
    if(uses_pointer)
    {
      strm << "&";
      implicitTemplateInstance (strm, lam, typeId);
      strm << "::fn" << id;
      comma = true;
    }

    if (ccount)
      {
        unsigned count = 0;
      
        for (paramListT::const_iterator iter = lam.context->begin()
               ; iter != lam.context->end()
               ; ++iter)
          {
            ++count;

            if(comma) strm << ", ";
            strm << "p" << count;
            comma = true;
          }
      }

    strm << ");\n  }\n};\n";
  }

  struct MatchInitialiser
  {
    const stringT &mTarget;
    bool mByRef;

    MatchInitialiser (const stringT &target
		      , bool byRef)
      : mTarget (target)
      , mByRef (byRef)
    {
    }

    bool operator() (const boost::shared_ptr<paramT> &pp)
    {
      return (mTarget == (*pp->initialiser)) && (mByRef == pp->byReference);
    }
  };

  bool CodeGen::contextRef (std::stringstream& outfile
			    , std::auto_ptr<stringT> strp
			    , bool byRef)
  {
    // Handles context by value and by reference

    lambdaT &lam (*mStack.top().mLamp);

    if (lam.explicitContext)
      {
	cerr
	  << "__ctx or __ref expression in lambda with explicit context\n";

	return false;
      }

    if (!lam.context.get())
      {
	lam.context.reset (new paramListT);
      }

    paramListT &context (*lam.context);

    paramListT::iterator param;

    param = std::find_if (context.begin()
			  , context.end()
			  , MatchInitialiser (*strp, byRef));

    if (param == context.end())
      {
	int ccount = context.size() + 1;

	char newName[128];

	sprintf (newName, "mContext%d", ccount);

	char pseudoType[128];

	sprintf (pseudoType, "T%d", ccount);

	boost::shared_ptr<paramT>
	  newParamPtr (new paramT (makestr(pseudoType)
				   , makestr (newName)
				   , strp.release()
                                   , ccount
				   , byRef));

	context.push_back (newParamPtr);

	param = context.end();
	--param;
      }

    outfile << contextVarExpression(**param);

    return true;
  }

  bool CodeGen::startLambda (std::stringstream*& f
			     , int openBracePos
			     , std::auto_ptr<lambdaT> lamp)
  {
    LambdaContext ctx(f, openBracePos, lamp);
    ctx.mFileName = mCurFileName;
    
    mStack.push (ctx);
    f            = new std::stringstream;
    mCurFileName = bodyFileName(++mLambdaCount);
    
    return true;
  }

  CodeGen::LambdaMap::iterator CodeGen::lambdaGroup (const LambdaCategory& typeId)
  {
    LambdaMap::iterator mapIter = mLambdaMap.find (typeId);

    if (mapIter == mLambdaMap.end())
      {
	pair<LambdaMap::iterator, bool> inserted
	  = mLambdaMap.insert (std::make_pair (typeId, LambdaGroup()));
	assert (inserted.second);
	mapIter = inserted.first;
	//
      }

    return mapIter;
  }

  unsigned CodeGen::categorysize(const LambdaCategory& typeId) const
  {
    LambdaMap::const_iterator mapIter = mLambdaMap.find (typeId);
    if(mapIter == mLambdaMap.end()) return 0;
    unsigned res = mapIter->second.mLambdaList.size();
    /*
    std::stringstream name;
    implTemplateName(name, typeId);
    fprintf(stderr, "cat size(%s)=%u\n", name.str().c_str(), res);
    */
    return res;
  };
  const std::string CodeGen::findFirstFunc(const LambdaCategory& typeId) const
  {
    LambdaMap::const_iterator mapIter = mLambdaMap.find (typeId);
    if(mapIter == mLambdaMap.end()) return "";
    const LambdaContext& ctx = mapIter->second.mLambdaList.front();
    int fnum = ctx.mLambdaNumber;
    
    std::stringstream tmp;
    tmp << "fn" << fnum;
    return tmp.str();
  }

  
  bool CodeGen::endLambda (std::stringstream*& f)
  {
    *f << "\n"; // Ensure newline termination
    
    LambdaFileContents[mCurFileName] = f->str();
    
    delete f;
    
    f            = mStack.top().mPrevFile;
    mCurFileName = mStack.top().mFileName;

    LambdaContext& ctx (mStack.top());

    LambdaCategory typeId;
     typeId.ccount = count (ctx.mLamp->context.get());
     typeId.pcount = count (ctx.mLamp->params.get());
     typeId.quals  = ctx.mLamp->qual_str.get() ? *ctx.mLamp->qual_str.get() : "";

    LambdaMap::iterator mapIter = lambdaGroup (typeId);

    mapIter->second.mLambdaList.push_back (ctx);

    declareMemberFn (*mapIter->second.mImplStream
		     , ctx.mLambdaNumber
		     , typeId);

    std::stringstream temp;
    
    if (ctx.mLamp->explicitContext)
      {
	explicitConstructor (temp, ctx.mLambdaNumber, *ctx.mLamp);
      }

    else
      {
	writeGenerator (mGeneratorStream, ctx.mLambdaNumber, *ctx.mLamp);
	implicitConstructor (temp, ctx.mLambdaNumber, *ctx.mLamp);
      }
      
    *f << temp.str();

    mStack.pop();
    return true;
  }

  extern "C" { extern FILE* yyout; }
  
  void CodeGen::spitItOut (const std::stringstream& program)
  {
    mImplStream
      << "//"
      << "\n// clamp generated lambda templates"
      << "\n// parser " << clamp_parser_rcsid
      << "\n// code generator " << clamp_codegen_rcsid
      << "\n//\n\n"
      ;

    mImplStream << "namespace {\n";
    for (LambdaMap::const_iterator iter = mLambdaMap.begin()
	   ; iter != mLambdaMap.end()
	   ; ++iter)
      {
	startExplicitImpl (mImplStream, iter->first);
	mImplStream << iter->second.mImplStream->str();
	
	if(true) // iter->first.ccount == 0)
	{
          implTemplateName (mImplStream, iter->first);
          mImplStream << "() {}\n";
          
          implTemplateName (mImplStream, iter->first);
          mImplStream << "& operator= (";
          implTemplateName (mImplStream, iter->first);
          mImplStream << " const &b) {\n";
          
          mImplStream << "mFn = b.mFn;\n";
          
          for(unsigned a=0; a<iter->first.ccount; ++a)
                mImplStream << "mContext" << a+1 << " = b.mContext" << a+1 << ";\n";
          
          mImplStream << "return *this;\n";
          
          mImplStream << "}\n";
	}
	else
	{
          mImplStream << "\nprivate:\n  ";

          // Make the assignment operator private, since it wouldn't
          // handle any reference member variables

          implTemplateName (mImplStream, iter->first);

          mImplStream << " operator= (";

          implTemplateName (mImplStream, iter->first);

          mImplStream << " const &);\n";
        }
        mImplStream << "};\n";
      }

    mImplStream << "\n" << mGeneratorStream.str() << "\n";

    mImplStream << "} /* anonymous namespace */\n";

    
    std::stringstream members;
    
    members << mImplStream.str();
    members << "\n";
    
#if 0
    members << "#line 0 \"" << infn << "\"\n";
#endif
    
    members << program.str();

    for (LambdaMap::const_iterator group = mLambdaMap.begin()
	   ; group != mLambdaMap.end()
	   ; ++group)
      {
	const LambdaList &list (group->second.mLambdaList);

	for (LambdaList::const_iterator iter = list.begin()
	       ; iter != list.end()
	       ; ++iter)
	  {
	    defineMemberFn (members
			    , iter->mLambdaNumber
			    , group->first
			    , *iter->mLamp);
	  }
      }
    
    fprintf(yyout, "%s\n", members.str().c_str());
  }
}
